# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""A wrapper class that executes Docker CLI commands via subprocess."""
import subprocess

import click


class KafkaCLIWrapper:
    """A utility class for interacting with Kafka using command-line tools.

    This class provides methods to consume messages from a
    Kafka topic using the Kafka console consumer.
    """

    @classmethod
    def consume_kafka_topic(cls, topic_name: str, from_beginning: bool) -> None:
        """Consume messages from a Kafka topic using the Kafka console consumer.

        Args:
            topic_name: The name of the Kafka topic to consume
            messages from.

            from_beginning: If True, start consuming from the
            beginning of the topic. If False, start consuming from
            the current offset.

        Raises:
            subprocess.CalledProcessError: If there is an error running
            the Kafka viewer.

        Example:
            To consume messages from a Kafka topic 'my-topic' from the beginning

            >>> KafkaCLIWrapper.consume_kafka_topic('my-topic', True)
        """
        if from_beginning:
            command = (
                "docker exec -it broker kafka-console-consumer"
                " --bootstrap-server localhost:9092 --topic"
                f" {topic_name} --from-beginning"
            )
        else:
            command = (
                "docker exec -it broker kafka-console-consumer"
                f" --bootstrap-server localhost:9092 --topic {topic_name}"
            )
        try:
            with subprocess.Popen(
                command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                bufsize=1,  # Line-buffered
                universal_newlines=True,
            ) as process:
                if process.stdout is not None:
                    for line in process.stdout:
                        print(f"{line.strip()}\r")

            process.wait()
        except subprocess.CalledProcessError as ex:
            print(f"Error running Kafka viewer: {ex}")
            print(f"Command output: {ex.output}")


@click.command()
@click.argument("dataset")
@click.option(
    "-b",
    "--from-beginning",
    is_flag=True,
    help="If messages should be streamed from the start",
)
def stream(dataset: str, from_beginning: bool) -> None:
    """Stream messages from a dataset."""
    KafkaCLIWrapper.consume_kafka_topic(dataset, from_beginning=from_beginning)



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Package information for the Aineko package."""


__version__ = "0.2.4"
__author__ = "Convex Labs Engineering"

import logging

from aineko.core.config_loader import ConfigLoader
from aineko.core.dataset import (
    DatasetConsumer,
    DatasetProducer,
    FakeDatasetConsumer,
    FakeDatasetProducer,
)
from aineko.core.node import AbstractNode
from aineko.core.runner import Runner


def _setup_logging() -> None:
    """Setup logging for the application."""
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s.%(msecs)d - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )


_setup_logging()



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Aineko command line interface."""
import click

from aineko import __version__
from aineko.cli.create_pipeline import create
from aineko.cli.docker_cli_wrapper import service
from aineko.cli.kafka_cli_wrapper import stream
from aineko.cli.run import run
from aineko.cli.visualize import visualize


@click.group("aineko")
@click.version_option(__version__)
def cli() -> None:
    """Aineko CLI."""
    pass


cli.add_command(create)
cli.add_command(run)
cli.add_command(service)
cli.add_command(stream)
cli.add_command(visualize)

if __name__ == "__main__":
    cli(obj={})



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Example file showing how to create nodes."""

import time
from typing import Optional

from aineko.core.node import AbstractNode


class MySequencerNode(AbstractNode):
    """Example node."""

    def _pre_loop_hook(self, params: Optional[dict] = None) -> None:
        """Optional; used to initialize node state."""
        self.current_val = params.get("initial_state", 0)

    def _execute(self, params: Optional[dict] = None) -> Optional[bool]:
        """Required; function repeatedly executes.

        Accesses inputs via `self.consumer`, and outputs via
        `self.producer`.
        Logs can be sent via the `self.log` method.
        """
        self.current_val += params.get("increment", 1)
        time.sleep(5)
        self.producers["test_sequence"].produce(self.current_val)


class PrintInput(AbstractNode):
    """Prints input."""

    def _execute(self, params: Optional[dict] = None):
        for dataset, consumer in self.consumers.items():
            msg = consumer.consume(how="next")
            if msg is None:
                continue
            print(
                f"Node {params['name']} received message: {msg['message']} from {dataset}"
            )


class MySumNode(AbstractNode):
    """Example node."""

    def _pre_loop_hook(self, params: Optional[dict] = None) -> None:
        """Optional; used to initialize node state."""
        self.state = params.get("initial_state", 0)

    def _execute(self, params: Optional[dict] = None) -> Optional[bool]:
        """Required; function repeatedly executes.

        Accesses inputs via `self.consumer`, and outputs via
        `self.producer`.
        Logs can be sent via the `self.log` method.
        """
        msg = self.consumers["test_sequence"].consume(how="next")
        if msg is None:
            return
        self.log(f"Received: {msg['message']}. Adding {params['increment']}...")
        self.state = int(msg["message"]) + int(params["increment"])
        self.producers["test_sum"].produce(self.state)



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Module for managing datasets.

Datasets are Kafka topics that are used to
communicate between nodes in the pipeline. The dataset module provides
a wrapper for the Kafka consumer and producer objects.

Note:
e.g. message:
{
    # Inferred metadata
    "timestamp": "2020-01-01 00:00:00"
    "dataset": "dataset_1",
    "source_node": "node_1",

    # User defined message
    "message": {...},
}
"""
import datetime
import json
import logging
from typing import Any, Dict, Optional

from confluent_kafka import Consumer, Message, Producer  # type: ignore

from aineko.config import AINEKO_CONFIG, DEFAULT_KAFKA_CONFIG

logger = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class DatasetConsumer:
    """Wrapper class for Kafka consumer object.

    DatasetConsumer objects are designed to consume messages from a single
    dataset and will consume the next unconsumed message in the queue.

    When accessing kafka topics, prefixes will automatically be added to the
    dataset name as part of namespacing. For datasets defined in the pipeline
    config, `has_pipeline_prefix` will be set to `True`, so a dataset named
    `my_dataset` will point to a topic named `my_pipeline.my_dataset`.

    Optionally, a custom prefix can be provided that will apply to all datasets.
    In the above example, if the prefix is set to `test`, the topic name will
    be `test.my_pipeline.my_dataset`.

    Args:
        dataset_name: name of the dataset
        node_name: name of the node that is consuming the dataset
        pipeline_name: name of the pipeline
        dataset_config: dataset config
        broker: broker to connect to (ip and port: "54.88.142.21:9092")
        prefix: prefix for topic name (<prefix>.<dataset_name>)
        has_pipeline_prefix: whether the dataset name has pipeline name prefix

    Attributes:
        consumer: Kafka consumer object

    Methods:
        consume: reads a message from the dataset
    """

    def __init__(
        self,
        dataset_name: str,
        node_name: str,
        pipeline_name: str,
        dataset_config: Dict[str, Any],
        bootstrap_servers: Optional[str] = None,
        prefix: Optional[str] = None,
        has_pipeline_prefix: bool = False,
    ):
        """Initialize the consumer."""
        self.pipeline_name = pipeline_name
        self.kafka_config = DEFAULT_KAFKA_CONFIG
        self.prefix = prefix
        self.has_pipeline_prefix = has_pipeline_prefix

        consumer_config = self.kafka_config.get("CONSUMER_CONFIG")
        # Overwrite bootstrap server with broker if provided
        if bootstrap_servers:
            consumer_config["bootstrap.servers"] = bootstrap_servers

        # Override default config with dataset specific config
        for param, value in dataset_config.get("params", {}).items():
            consumer_config[param] = value

        topic_name = dataset_name
        if has_pipeline_prefix:
            topic_name = f"{pipeline_name}.{dataset_name}"

        if self.prefix:
            consumer_config[
                "group.id"
            ] = f"{prefix}.{pipeline_name}.{node_name}"
            self.consumer = Consumer(consumer_config)
            self.consumer.subscribe([f"{prefix}.{topic_name}"])

        else:
            consumer_config["group.id"] = f"{pipeline_name}.{node_name}"
            self.consumer = Consumer(consumer_config)
            self.consumer.subscribe([topic_name])

        self.topic_name = topic_name

    @staticmethod
    def _validate_message(
        message: Optional[Message] = None,
    ) -> Optional[dict]:
        """Checks if a message is valid and converts it to appropriate format.

        Args:
            message: message to check

        Returns:
            message: message if valid, None if not
        """
        # Check if message is valid
        if message is None or message.value() is None:
            return None

        # Check if message is an error
        if message.error():
            logger.error(str(message.error()))
            return None

        # Convert message to dict
        message = message.value()
        if isinstance(message, bytes):
            message = message.decode("utf-8")
        return json.loads(message)

    def consume(
        self,
        how: str = "next",
        timeout: Optional[int] = None,
    ) -> Optional[dict]:
        """Reads a message from the dataset.

        Args:
            how: how to read the message
                "next": read the next message in the queue

        Returns:
            msg: message from the dataset
        """
        timeout = timeout or self.kafka_config.get("CONSUMER_TIMEOUT")
        if how == "next":
            # next unread message from queue
            message = self._validate_message(
                self.consumer.poll(timeout=timeout)
            )
        else:
            raise ValueError(f"Invalid how: {how}. Expected 'next'.")

        return message

    def consume_all(self, end_message: str | bool = False) -> list:
        """Reads all messages from the dataset until a specific one is found.

        Args:
            end_message: Message to trigger the completion of consumption

        Returns:
            list: list of messages from the dataset
        """
        messages = []
        while True:
            message = self.consume()
            if message is None:
                continue
            if message["message"] == end_message:
                break
            messages.append(message)
        return messages


class DatasetProducer:
    """Wrapper class for Kafka producer object.

    See DatasetConsumer for prefix rules.

    Args:
        dataset_name: dataset name
        node_name: name of the node that is producing the message
        pipeline_name: name of the pipeline
        dataset_config: dataset config
        prefix: prefix for topic name (<prefix>.<dataset_name>)
        has_pipeline_prefix: whether the dataset name has pipeline name prefix

    Attributes:
        producer: Kafka producer object

    Methods:
        produce: produce a message to the dataset
    """

    def __init__(
        self,
        dataset_name: str,
        node_name: str,
        pipeline_name: str,
        dataset_config: Dict[str, Any],
        prefix: Optional[str] = None,
        has_pipeline_prefix: bool = False,
    ):
        """Initialize the producer."""
        self.source_pipeline = pipeline_name
        self.dataset = dataset_name
        self.source_node = node_name
        self.prefix = prefix
        self.has_pipeline_prefix = has_pipeline_prefix

        # Create topic name based on prefix rules
        topic_name = dataset_name
        if has_pipeline_prefix:
            topic_name = f"{pipeline_name}.{topic_name}"
        if prefix:
            topic_name = f"{prefix}.{topic_name}"
        self.topic_name = topic_name

        # Assign kafka config
        self.kafka_config = DEFAULT_KAFKA_CONFIG

        # Set producer parameters
        producer_config = self.kafka_config.get("PRODUCER_CONFIG")

        # Override default config with dataset specific config
        if "params" in dataset_config:
            for param in self.kafka_config.get("PRODUCER_OVERRIDABLES"):
                if param in dataset_config["params"]:
                    producer_config[param] = dataset_config["params"][param]

        # Create producer
        self.producer = Producer(producer_config)

    @staticmethod
    def _delivery_report(err: Any, message: Message) -> None:
        """Called once for each message produced to indicate delivery result.

        Triggered by poll() or flush().

        Args:
            err: error message
            message: message object from Kafka
        """
        if err is not None:
            logger.error("Message %s delivery failed: %s", message, err)

    def produce(self, message: dict, key: Optional[str] = None) -> None:
        """Produce a message to the dataset.

        Args:
            message: message to produce to the dataset
            key: key to use for the message
        """
        message = {
            "timestamp": datetime.datetime.now().strftime(
                AINEKO_CONFIG.get("MSG_TIMESTAMP_FORMAT")
            ),
            "dataset": self.dataset,
            "source_pipeline": self.source_pipeline,
            "source_node": self.source_node,
            "message": message,
        }
        self.producer.poll(0)

        key_bytes = str(key).encode("utf-8") if key is not None else None

        self.producer.produce(
            topic=self.topic_name,
            key=key_bytes,
            value=json.dumps(message).encode("utf-8"),
            callback=self._delivery_report,
        )
        self.producer.flush()


# pylint: enable=too-few-public-methods
# pylint: disable=unused-argument
class FakeDatasetConsumer:
    """Fake dataset consumer for testing purposes.

    The class will store in its state the list of values to feed the node,
    and pop each value everytime the consume method is called.

    Args:
        dataset_name: name of the dataset
        node_name: name of the node that is consuming the dataset
        values: list of values to feed the node

    Attributes:
        dataset_name (str): name of the dataset
        node_name (str): name of the node that is consuming the dataset
        values (list): list of mock data values to feed the node
        empty (bool): True if the list of values is empty, False otherwise
    """

    def __init__(self, dataset_name: str, node_name: str, values: list):
        """Initialize the consumer."""
        self.dataset_name = dataset_name
        self.node_name = node_name
        self.values = values
        self.empty = False

    def consume(
        self,
        how: str = "next",
        timeout: Optional[int] = None,
    ) -> Optional[dict]:
        """Reads a message from the dataset.

        Args:
            how: how to read the message
                "next": read the next message in the queue

        Returns:
            next value in self.values

        Raises:
            ValueError: if how is not "next"
        """
        if how == "next":
            remaining = len(self.values)
            if remaining > 0:
                if remaining == 1:
                    self.empty = True
                return {
                    "timestamp": datetime.datetime.now().strftime(
                        AINEKO_CONFIG.get("MSG_TIMESTAMP_FORMAT")
                    ),
                    "message": self.values.pop(0),
                    "source_node": "test",
                    "source_pipeline": "test",
                }
            else:
                return None
        raise ValueError(f"Invalid how: {how}. Expected 'next'.")


class FakeDatasetProducer:
    """Fake dataset producer for testing purposes.

    The class will store in its state the list of values produced
    by the node.

    Args:
        dataset_name: name of the dataset
        node_name: name of the node that is producing the dataset

    Attributes:
        dataset_name (str): name of the dataset
        node_name (str): name of the node that is producing the dataset
        values (list): list of mock data values produced by the node
    """

    def __init__(self, dataset_name: str, node_name: str):
        """Initialize the producer."""
        self.dataset_name = dataset_name
        self.node_name = node_name
        self.values = []  # type: ignore

    def produce(self, message: Any) -> None:
        """Stores message in self.values.

        Args:
            message: message to produce.
        """
        self.values.append(message)



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Example file showing unit testing."""

from lib.nodes import MySumNode


def test_mynode():
    """Unit test for MySumNode."""

    mynode = MySumNode(test=True)
    mynode.setup_test(
        inputs={
            "test_sequence": [1, 2, 3]
        },  # input a list of elements to be consumed
        outputs=["test_sum"],  # list of dataset names that are produced to
        params={"initial_state": 0, "increment": 1},
    )
    outputs = mynode.run_test()
    assert outputs["test_sum"] == [2, 3, 4]
    assert mynode.state == 4



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Import utilities."""
import importlib
import inspect
from typing import Any, Optional


def import_from_string(  # type: ignore[no-untyped-def]
    attr: str,
    kind: str,
    reqd_params: Optional[list] = None,
    ret_type: Optional[Any] = None,
):
    """Get a function from a string.

    Args:
        attr: String to import (e.g. ...module.function or
              ...module.Class)
        kind: Kind of object to import (function or class)
        reqd_params: Required parameters of a function or class
        ret_type: Return type of a function (ignored if kind is class)

    Returns:
        imported_func: Imported function

    Raises:
        ValueError: If kind is not a valid kind
        ValueError: If attr is not in the correct format
        ValueError: If attr is not a function
        ValueError: If attr is not a class
    """
    # Check if kind is valid
    if kind not in ["function", "class"]:
        raise ValueError(
            f"Kind {kind} is not a valid kind. "
            "Valid kinds are 'function' and 'class'."
        )

    # Check if attribute is in the correct format
    if "." not in attr:
        raise ValueError(
            f"Attribute {attr} must be in the format "
            "...module.function or ...module.Class"
        )

    # Import attribute from module
    module_name, attr_name = attr.rsplit(".", 1)
    module = importlib.import_module(module_name)
    imported_attr = getattr(module, attr_name)

    # Validate attribute
    if kind == "function":
        if not inspect.isfunction(imported_attr):
            raise ValueError(
                f"Attribute {imported_attr} is {type(imported_attr)}. "
                "Expected a function."
            )

    elif kind == "class":
        if not inspect.isclass(imported_attr):
            raise ValueError(
                f"Attribute {imported_attr} is {type(imported_attr)}. "
                "Expected a class."
            )

    # Check if required parameters are present
    if isinstance(reqd_params, list):
        for param in reqd_params:
            if param not in inspect.signature(imported_attr).parameters:
                raise ValueError(
                    f"Imported function {imported_attr} "
                    f"does not have a {param} parameter."
                )

    # Check if returned type is correct
    if isinstance(ret_type, str):
        imported_attr_type = inspect.signature(imported_attr).return_annotation
        if imported_attr_type != ret_type:
            raise ValueError(
                f"Imported function {imported_attr} returns "
                f"{imported_attr_type}. Expected {ret_type}."
            )

    return imported_attr



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Utils to create a aineko pipeline from existing templates."""

import os

import click
from cookiecutter.main import cookiecutter  # type: ignore


@click.command()
@click.option(
    "-d",
    "--deployment-config",
    is_flag=True,
    help="Include deploy.yml that facilitates deployment of pipelines.",
)
@click.option(
    "-o",
    "--output-dir",
    default=".",
    help="Directory to create pipeline in. Defaults to current directory.",
)
@click.option(
    "-n",
    "--no-input",
    is_flag=True,
    help="Do not prompt for parameters and use defaults.",
)
def create(deployment_config: bool, output_dir: str, no_input: bool) -> None:
    """Creates boilerplate code and config required for an Aineko pipeline.\f

    Args:
        deployment_config: If True, include a deploy file when generating, else
        do not include.
        output_dir: Directory to create pipeline in. Defaults to current.
        no_input: If True, do not prompt for parameters and use defaults.
    """
    script_directory = os.path.dirname(os.path.abspath(__file__))
    # assumes templates is a folder that is one-level up
    templates_directory = f"{os.path.dirname(script_directory)}/templates"
    # Only one type of pipeline, hence this argument is no-arg
    cookiecutter(
        f"{templates_directory}/first_aineko_pipeline",
        extra_context={
            "_deployment_config": deployment_config,
        },
        output_dir=output_dir,
        no_input=no_input,
    )



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""A series of cookiecutter pre-codegen hooks to validate user input.

These code will run after user-input is provided and before files are
generated.

See: https://cookiecutter.readthedocs.io/en/1.7.2/advanced/hooks.html
"""
import re
import sys

# Color codes for logging
ERROR = "\x1b[1;31m[ERROR]: "
TERMINATOR = "\x1b[0m"

MODULE_REGEX = r"^[_a-zA-Z][_a-zA-Z0-9]+$"

project_slug = "{{ cookiecutter.project_slug }}"
pipeline_slug = "{{ cookiecutter.pipeline_slug }}"


class AinekoPathValidationException(Exception):
    """Exception raised for invalid aineko-core repo paths.

    This exception is thrown when the user-provided path does not point
    to a valid aineko-core repository.
    """


def validate_slug(slug: str) -> None:
    """Validate a slug to ensure compliance with specified rules.

    Args:
        slug: The slug to be validated.

    Raises:
        ValueError: Raised if the slug contains invalid characters.
    """
    if not re.match(r"^[a-z_-]+$", slug):
        raise ValueError(
            f"[{slug}] Slug can only contain lowercase letters, dashes,"
            " and underscores."
        )


try:
    validate_slug(project_slug)
    validate_slug(pipeline_slug)


except ValueError as ex:
    print(ERROR + str(ex) + TERMINATOR)
    sys.exit(1)

except AinekoPathValidationException as ex:
    print(ERROR + str(ex) + TERMINATOR)
    sys.exit(1)



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Example file setting default configs for project."""

from aineko.config import BaseConfig


# pylint: disable=invalid-name
class DEFAULT_CONFIG(BaseConfig):
    """Default configs for project."""

    NUM_CPUS = 0.5



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Functions for validation checks used in pydantic models."""

import re


def check_power_of_2(x: int) -> int:
    """Check if the given integer is a power of 2.

    Args:
        x: integer to check

    Returns:
        The input integer if it is a power of 2.

    Raises:
        ValueError: if the input integer is not a power of 2.
    """
    if not ((x & (x - 1) == 0) and x != 0):
        raise ValueError("Value must be a power of 2")
    return x


def check_semver(x: str) -> str:
    """Check if the given string is a semantic version.

    Regex taken from:
    https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string

    Args:
        x: string to check

    Returns:
        The input string if it is a semantic version.

    Raises:
        ValueError: if the input string is not a semantic version.
    """
    regex = (
        r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)"
        r"(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
        r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))"
        r"?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
    )
    if re.search(regex, x) is None:
        raise ValueError("Version must be in the form `1.2.3`")
    return x



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Schema models for deployment configuration specified in deploy.yml.

DeploymentConfig is the schema for a user written deploy.yml file. This format
is more compact, and allows for defining re-usable config under the `defaults`
and `pipelines` keys.

FullDeploymentConfig is the schema for the full deployment configuration, which
is the machine-readable version of DeploymentConfig. All re-usable config is
explicitly injected into the config so all pipelines are explicitly defined.
This config represents the source of truth for all deployments of aineko
pipelines.
"""

from typing import Dict, Optional

from pydantic import BaseModel, validator

from aineko.models.deploy_config_schema_internal import (
    Environment,
    GenericPipeline,
    ParameterizableDefaults,
)


class DeploymentConfig(BaseModel, extra="forbid"):
    """User deployment configuration (Schema for deploy.yml)."""

    version: str
    defaults: Optional[ParameterizableDefaults]
    pipelines: Dict[str, GenericPipeline]
    environments: Dict[str, Environment]

    @validator("version")
    def semver(cls, v: str) -> str:  # pylint: disable=no-self-argument
        """Validates that versioning follow semver convention."""
        if len(v.split(".")) != 3:
            raise ValueError("Version must be in the form `1.2.3`")
        return v


class FullDeploymentConfig(BaseModel):
    """Full deployment configuration (Schema for deploy.yml)."""

    version: str
    environments: Dict[str, Environment]

    @validator("version")
    def semver(cls, v: str) -> str:  # pylint: disable=no-self-argument
        """Validates that versioning follow semver convention."""
        if len(v.split(".")) != 3:
            raise ValueError("Version must be in the form `1.2.3`")
        return v



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Internal models for pipeline config validation."""
from typing import Optional

from pydantic import BaseModel, Field


class Config(BaseModel):
    """Config model."""

    class Pipeline(BaseModel):
        """Pipeline model."""

        class Dataset(BaseModel):
            """Dataset model."""

            type: str
            params: Optional[dict]

        class Node(BaseModel):
            """Node model."""

            class_name: str = Field(..., alias="class")
            node_params: Optional[dict]
            node_settings: Optional[dict]
            inputs: Optional[list]
            outputs: Optional[list]

        name: str
        default_node_settings: Optional[dict]
        nodes: dict[str, Node]
        datasets: dict[str, Dataset]

    pipeline: Pipeline



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Submodule that handles the running of a pipeline from config."""
import logging
import time
from typing import Optional

import ray
from confluent_kafka.admin import AdminClient, NewTopic  # type: ignore

from aineko.config import (
    AINEKO_CONFIG,
    DEFAULT_KAFKA_CONFIG,
    NODE_MANAGER_CONFIG,
)
from aineko.core.config_loader import ConfigLoader
from aineko.core.node import PoisonPill
from aineko.utils import imports

logger = logging.getLogger(__name__)


class Runner:
    """Runs the pipeline described in the config.

    Args:
        pipeline_config_file (str): Path to pipeline config file
        pipeline_name (str): Name of the pipeline
        kafka_config (dict): Config for kafka broker
        dataset_prefix (Optional[str]): Prefix for dataset names.
        Kafka topics will be called <prefix>.<pipeline>.<dataset_name>.

    Attributes:
        pipeline_config_file (str): Path to pipeline config file
        pipeline_name (str): Name of the pipeline, overrides pipeline config
        kafka_config (dict): Config for kafka broker
        pipeline_name (str): Name of the pipeline, loaded from config
        dataset_prefix (Optional[str]): Prefix for dataset names
    """

    def __init__(
        self,
        pipeline_config_file: str,
        pipeline_name: Optional[str] = None,
        kafka_config: dict = DEFAULT_KAFKA_CONFIG.get("BROKER_CONFIG"),
        metrics_export_port: int = AINEKO_CONFIG.get("RAY_METRICS_PORT"),
        dataset_prefix: Optional[str] = None,
    ):
        """Initializes the runner class."""
        self.pipeline_config_file = pipeline_config_file
        self.kafka_config = kafka_config
        self.metrics_export_port = metrics_export_port
        self.pipeline_name = pipeline_name
        self.dataset_prefix = dataset_prefix or ""

    def run(self) -> None:
        """Runs the pipeline.

        Step 1: Load config for pipeline

        Step 2: Set up datasets

        Step 3: Set up PoisonPill node that is available to all nodes

        Step 4: Set up nodes (including Node Manager)
        """
        # Load pipeline config
        pipeline_config = self.load_pipeline_config()
        self.pipeline_name = self.pipeline_name or pipeline_config["name"]

        # Create the necessary datasets
        self.prepare_datasets(
            config=pipeline_config["datasets"],
            user_dataset_prefix=self.pipeline_name,
        )

        # Initialize ray cluster
        ray.shutdown()
        ray.init(
            namespace=self.pipeline_name,
            _metrics_export_port=self.metrics_export_port,
        )

        # Create poison pill actor
        poison_pill = ray.remote(PoisonPill).remote()

        # Add Node Manager to pipeline config
        pipeline_config["nodes"][
            NODE_MANAGER_CONFIG.get("NAME")
        ] = NODE_MANAGER_CONFIG.get("NODE_CONFIG")

        # Create each node (actor)
        results = self.prepare_nodes(
            pipeline_config=pipeline_config,
            poison_pill=poison_pill,  # type: ignore
        )

        ray.get(results)

    def load_pipeline_config(self) -> dict:
        """Loads the config for a given pipeline.

        Returns:
            pipeline config
        """
        config = ConfigLoader(
            pipeline_config_file=self.pipeline_config_file,
        ).load_config()

        return config["pipeline"]

    def prepare_datasets(
        self, config: dict, user_dataset_prefix: Optional[str] = None
    ) -> bool:
        """Creates the required datasets for a given pipeline.

        Datasets can be configured using the `params` key, using config keys
        found in: https://kafka.apache.org/documentation.html#topicconfigs

        Args:
            config: dataset configuration found in pipeline config
            Should follow the schema
                {
                    "dataset_name": {
                        "type": str ("kafka_stream"),
                        "params": dict
                }
            user_dataset_prefix: prefix only for datasets defined by the user.
            i.e. <prefix>.<user_dataset_prefix>.<dataset_name>

        Returns:
            True if successful

        Raises:
            ValueError: if dataset "logging" is defined in the catalog
        """
        # Connect to kafka cluster
        kafka_client = AdminClient(self.kafka_config)

        # Add prefix to user defined datasets
        if user_dataset_prefix:
            config = {
                f"{user_dataset_prefix}.{dataset_name}": dataset_config
                for dataset_name, dataset_config in config.items()
            }

        # Fail if reserved dataset names are defined in catalog
        for reserved_dataset in DEFAULT_KAFKA_CONFIG.get("DATASETS"):
            if reserved_dataset in config:
                raise ValueError(
                    f"Unable to create dataset `{reserved_dataset}`. "
                    "Reserved for internal use."
                )

        # Add logging dataset to catalog
        config[DEFAULT_KAFKA_CONFIG.get("LOGGING_DATASET")] = {
            "type": AINEKO_CONFIG.get("KAFKA_STREAM_TYPE"),
            "params": DEFAULT_KAFKA_CONFIG.get("DATASET_PARAMS"),
        }

        # Create all dataset defined in the catalog
        dataset_list = []
        for dataset_name, dataset_config in config.items():
            logger.info(
                "Creating dataset: %s: %s", dataset_name, dataset_config
            )
            # Create dataset for kafka streams
            if dataset_config["type"] == AINEKO_CONFIG.get("KAFKA_STREAM_TYPE"):
                # User defined
                dataset_params = {
                    **DEFAULT_KAFKA_CONFIG.get("DATASET_PARAMS"),
                    **dataset_config.get("params", {}),
                }

                # Configure dataset
                if self.dataset_prefix:
                    topic_name = f"{self.dataset_prefix}.{dataset_name}"
                else:
                    topic_name = dataset_name

                new_dataset = NewTopic(
                    topic=topic_name,
                    num_partitions=dataset_params.get("num_partitions"),
                    replication_factor=dataset_params.get("replication_factor"),
                    config=dataset_params.get("config"),
                )

                # Add dataset to appropriate list
                dataset_list.append(new_dataset)

            else:
                raise ValueError(
                    "Unknown dataset type. Expected: "
                    f"{AINEKO_CONFIG.get('STREAM_TYPES')}."
                )

        # Create all configured datasets
        datasets = kafka_client.create_topics(dataset_list)

        # Block until all datasets finish creation
        cur_time = time.time()
        while True:
            if all(future.done() for future in datasets.values()):
                logger.info("All datasets created.")
                break
            if time.time() - cur_time > AINEKO_CONFIG.get(
                "DATASET_CREATION_TIMEOUT"
            ):
                raise TimeoutError(
                    "Timeout while creating Kafka datasets. "
                    "Please check your Kafka cluster."
                )

        return datasets

    def prepare_nodes(
        self, pipeline_config: dict, poison_pill: ray.actor.ActorHandle
    ) -> list:
        """Prepare actor handles for all nodes.

        Args:
            pipeline_config: pipeline configuration

        Returns:
            dict: mapping of node names to actor handles
            list: list of ray objects
        """
        # Collect all  actor futures
        results = []

        default_node_config = pipeline_config.get("default_node_settings", {})

        for node_name, node_config in pipeline_config["nodes"].items():
            # Initialize actor from specified class in config
            target_class = imports.import_from_string(
                attr=node_config["class"], kind="class"
            )
            actor_params = {
                **default_node_config,
                **node_config.get("node_settings", {}),
                "name": node_name,
                "namespace": self.pipeline_name,
            }

            wrapped_class = ray.remote(target_class)
            wrapped_class.options(**actor_params)
            actor_handle = wrapped_class.remote(poison_pill=poison_pill)

            # Setup input and output datasets
            outputs = node_config.get("outputs", [])
            actor_handle.setup_datasets.remote(
                inputs=node_config.get("inputs", None),
                outputs=outputs,
                datasets=pipeline_config["datasets"],
                node=node_name,
                pipeline=self.pipeline_name,
                has_pipeline_prefix=True,
            )

            # Setup internal datasets like logging, without pipeline prefix
            actor_handle.setup_datasets.remote(
                outputs=DEFAULT_KAFKA_CONFIG.get("DATASETS"),
                datasets=pipeline_config["datasets"],
                node=node_name,
                pipeline=self.pipeline_name,
            )

            # Create actor future (for execute method)
            results.append(
                actor_handle.execute.remote(
                    params=node_config.get("node_params", None)
                )
            )

        return results



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""A wrapper class that executes Docker CLI commands via subprocess."""
import subprocess
import sys
from typing import Optional

import click
from click.core import Context


class DockerCLIWrapper:
    """A wrapper class that executes Docker CLI commands via subprocess.

    This class provides methods to start, stop, and restart Docker services
    using docker-compose.

    Methods:
        start_service(cls) -> None:
            Start the Docker service.

        stop_service(cls) -> None:
            Stop the running Docker service.

        kill_service(cls) -> None:
            Kills the running Docker service.

        restart_service(cls) -> None:
            Restart the running Docker service.
    """

    _docker_compose_config = """
version: "3"
services:
  zookeeper:
    image: confluentinc/cp-zookeeper:7.3.0
    hostname: zookeeper
    container_name: zookeeper
    environment:
      ZOOKEEPER_CLIENT_PORT: 2181
      ZOOKEEPER_TICK_TIME: 2000

  broker:
    image: confluentinc/cp-kafka:7.3.0
    container_name: broker
    ports:
      - "9092:9092"
    depends_on:
      - zookeeper
    environment:
      KAFKA_BROKER_ID: 1
      KAFKA_ZOOKEEPER_CONNECT: "zookeeper:2181"
      KAFKA_LISTENER_SECURITY_PROTOCOL_MAP: PLAINTEXT:PLAINTEXT,PLAINTEXT_INTERNAL:PLAINTEXT
      KAFKA_ADVERTISED_LISTENERS: PLAINTEXT://localhost:9092,PLAINTEXT_INTERNAL://broker:29092
      KAFKA_OFFSETS_TOPIC_REPLICATION_FACTOR: 1
      KAFKA_TRANSACTION_STATE_LOG_MIN_ISR: 1
      KAFKA_TRANSACTION_STATE_LOG_REPLICATION_FACTOR: 1
"""

    def __init__(self, custom_config_path: Optional[str] = None) -> None:
        """Initialize DockerCLIWrapper class."""
        if custom_config_path:
            self._load_custom_config(custom_config_path)

    @classmethod
    def run_docker_command(cls, command: str) -> None:
        """Run a Docker CLI command.

        Args:
            command: The Docker CLI command to run.
        """
        try:
            output = subprocess.check_output(
                args=command,
                input=cls._docker_compose_config,
                shell=True,
                text=True,
                stderr=subprocess.STDOUT,
            )
            print(output)
        except subprocess.CalledProcessError as ex:
            print(f"Error: {ex}")
            print(f"Command Output: {ex.output}")

    @classmethod
    def start_service(cls) -> None:
        """Start the Docker service."""
        cls.run_docker_command("docker-compose -f - up -d")

    @classmethod
    def stop_service(cls) -> None:
        """Stop the running Docker service."""
        cls.run_docker_command("docker-compose -f - stop")

    @classmethod
    def kill_service(cls) -> None:
        """Kill the running Docker service."""
        cls.run_docker_command("docker-compose -f - down")

    @classmethod
    def restart_service(cls, hard: bool) -> None:
        """Restart the running Docker service."""
        if hard:
            cls.run_docker_command("docker-compose -f - down")
            cls.run_docker_command("docker-compose -f - up -d")
        else:
            cls.run_docker_command("docker-compose -f - restart")

    @classmethod
    def _load_custom_config(cls, custom_config_path: str) -> None:
        """Load a custom Docker Compose config file.

        Args:
            custom_config_path: Path to the custom Docker Compose config file.
        """
        try:
            with open(
                file=custom_config_path, mode="r", encoding="utf-8"
            ) as file:
                cls._docker_compose_config = file.read()
        except FileNotFoundError:
            print(
                f"FileNotFoundError: Custom config file `{custom_config_path}`"
                f" not found."
            )
            sys.exit(1)


@click.group()
@click.option(
    "-c",
    "--config",
    help="Path to the custom Docker Compose config file.",
)
@click.pass_context
def service(ctx: Context, config: Optional[str] = None) -> None:
    """Manage Aineko docker services (Kafka and Zookeeper containers)."""
    ctx.ensure_object(dict)
    ctx.obj["config"] = config


@service.command()
@click.pass_context
def start(ctx: Context) -> None:
    """Start Aineko docker services."""
    DockerCLIWrapper(ctx.obj["config"]).start_service()


@service.command()
@click.pass_context
def stop(ctx: Context) -> None:
    """Stop Aineko docker services."""
    DockerCLIWrapper(ctx.obj["config"]).stop_service()


@service.command()
@click.pass_context
def down(ctx: Context) -> None:
    """Kill Aineko docker services."""
    DockerCLIWrapper(ctx.obj["config"]).kill_service()


@service.command()
@click.option(
    "-h",
    "--hard",
    is_flag=True,
    help=(
        "Forces full restart Aineko docker services."
        "Clears data from Kafka cache."
    ),
)
@click.pass_context
def restart(ctx: Context, hard: bool = True) -> None:
    """Restart Aineko docker services."""
    DockerCLIWrapper(ctx.obj["config"]).restart_service(hard=hard)



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Module to run a pipeline from the command line."""
import logging
import time
import traceback
from typing import Optional

import click

from aineko.core.runner import Runner

logger = logging.getLogger(__name__)


@click.command()
@click.argument("pipeline-config-file")
@click.option(
    "-p",
    "--pipeline-name",
    help="Name of the pipeline to run.",
)
@click.option(
    "-r",
    "--retry",
    is_flag=True,
    help="Retry running the pipeline on failure.",
)
def run(
    pipeline_config_file: str,
    pipeline_name: Optional[str] = None,
    retry: bool = False,
) -> None:
    """Main function to run a pipeline from the command line.\f

    Args:
        pipeline_config_file: Path to the file containing the pipeline config
        pipeline_name: Name of the pipeline to run, overrides pipeline config
        retry: If true, retry running the pipeline on failure every 10 seconds
    """
    logger.info("Application is starting.")
    while True:
        try:
            runner = Runner(
                pipeline_config_file=pipeline_config_file,
                pipeline_name=pipeline_name,
            )
            runner.run()
        except Exception as e:  # pylint: disable=broad-except
            if not retry:
                raise e
            else:
                logger.error("Error running pipeline: %s", e)
                logger.debug(traceback.format_exc())
                time.sleep(10)
                continue
